#!/usr/bin/python
# -*- coding: utf-8 -*-
#
#Fichero manejador de ficheros
#Copyright (C) 2006  Inmaculada Luengo Merino, Néstor Arocha Rodríguez

#This file is part of Driza.
#
#Driza is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 2 of the License, or
#(at your option) any later version.
#
#Driza is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with Driza; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from rpy import *

class InterfazFichero:
	""" Interfaz entre datos y ficheros. Permite cargar o guardar un objeto contenedorregvar en un fichero. Se comunica con una interfazdatos. Es llamado por el gestor de ficheros """

	def __init__(self,interfazdatos):
		"""
		Inicializa el manejador de ficheros.
		Crea los objetos que se usaran para comprobar la extension de los archivos
		"""
		#VARIABLES PRIVADAS
		import re
		self.__id=interfazdatos
		#FIXME: El control de extensiones lo debe llevar el GestorFicheros, para asi advertir al usuario de lo que crea conveniente
		self.__driza=re.compile('.*\.driza$')
		self.__spss=re.compile('.*\.dat$')
		self.__txt=re.compile('.*\.txt$')

	#FUNCIONES PUBLICAS
	#TODO: Estudiar conveniencia del parametro tipo
	def save(self,fichero,datos,tipo=None):
		"""Guarda un fichero en el formato standard (Pickle)"""
		archivo = open(fichero, 'w')
		if self.__driza.match(fichero):
			import pickle
			pickle.dump(datos,archivo)
		archivo.close()

	#TODO: Estudiar conveniencia del parametro tipo
	def load(self,fichero,tipo=None):
		"""Carga un fichero con formato standard"""
		archivo = open(fichero, 'r')
		if self.__driza.match(fichero):
			import pickle
			datos=pickle.load(archivo) 
			archivo.close
			self.__id.insertarestado(datos)
		elif self.__spss.match(fichero):
			pass

	def importartexto(self,fichero,delimitadorreg,delimitadoratrib,cabeceras):
		"""Importa los datos de un fichero de texto"""
		if self.__txt.match(fichero):
			self.__id.borrartodo()
			datos=r.read_table(fichero,sep=delimitadoratrib,header=cabeceras)
			for pareja in datos.items():
				self.__id.appVar(pareja[0],"Real")
				self.__id[self.__id.getVar(pareja[0])]=pareja[1]


class ListaFicheros(list):
	"""
	Contenedor de la lista de ficheros. Funciona como cola LIFO, y hereda de list.
	El parametro pasado es la longitud de lista que se desea
	"""
	def __init__(self,n):
		#VARIABLES PRIVADAS
		self.__nelementos=n
		list.__init__(self)

	#FUNCIONES PUBLICAS
	def insert(self,elemento):
		"""Inserta un elemento, garantizando que siempre se va a colocar en el primer lugar"""
		if elemento in self:
			self.remove(elemento)
		list.insert(self,0,elemento)
		if len(self)>self.__nelementos:
			self.pop()

#TODO: Falta la clase Config tal cual
# Elementos a configurar:

	#tmpdir: Guarda la ruta del directorio de trabajo
	#rdir:   Guarda la ruta del ejecutable de R
	#vsplash: boolean que dice si aparece la ventana splash o no
	#lfichero: Lista de ficheros

class InterfazConfig:
	"""
	Clase encargada de gestionar la configuración. Permite leer y guardar la configuración en ficheros.
	"""
	def __init__(self):
		"""
		Funcion de inicialización del manejador de configuración 
		"""
		import os
		#VARIABLES PRIVADAS
		self.__fichero=os.environ["HOME"]+"/.driza" #Lugar de la configuración por defecto

	#FUNCIONES PÚBLICAS
	def __getitem__(self,fichero):
		"""
		Definicion del operador [], devuelve la configuración que se encuentra en un fichero, y None si no puede abrirlo
		"""
		import pickle
		try:
			archivo=open(fichero,"r")
		except: #Archivo nuevo
			return None
		else:
			config=pickle.load(archivo)
			archivo.close()
			return config



	def __setitem__(self, fichero, config):
		"""
		Guarda la configuración en el fichero indicado.
		"""
		import pickle
		if config["primeravez"]:
			config["primeravez"]=False
		archivo=open(fichero,"w")
		pickle.dump(config,archivo)
		archivo.close()
	
	def load(self):
		"""
		Carga la configuración del fichero de configuración del programa, devolviendo la configuración por defecto si no la encuentra
		"""
		config= self.__getitem__(self.__fichero)
		if not config:
			config=self.__default()
			self.save(config)
			config["primeravez"]=True
		return config

	def save(self,config):
		"""
		Guarda los valores pasados en el fichero de configuración del programa
		"""
		self.__setitem__(self.__fichero,config)

	#FUNCIONES PRIVADAS

	def __default(self):
		"""
		Devuelve la lista de parámetros por defecto
		"""
		listaficheros=ListaFicheros(5)
		return {"tmpdir":"/tmp","rdir":"/usr/bin/R","lfichero":listaficheros,"vsplash":True,"primeravez":False}

class AgenteConversion:
	"""Esta clase realiza una conversion de un tipo a otro, devolviendo la nueva variable y la lista de campos de registros ya transformados"""
	def __init__(self,interfazdatos):
		self.__interfazdatos=interfazdatos

	def __call__(self,variable,tipovariabledestino,metododeseado):
		from Driza import variables
		from Driza.conversion import CDefault
		if tipovariabledestino == "Entero":
			nuevavariable=variables.TEntero()
		elif tipovariabledestino == "Real":
			nuevavariable=variables.TReal()
		elif tipovariabledestino == "Booleano":
			nuevavariable=variables.TBooleano()
		elif tipovariabledestino == "Fecha":
			nuevavariable=variables.TFecha()
		nuevavariable.update(variable) #Copia de atributos de la variable
		lista=self.__interfazdatos.getCol(variable,filtrado=False)
		tipofuncion=eval(metododeseado)
		lafuncion=tipofuncion(variable,nuevavariable)
		nuevalista=map(lafuncion,lista)
		return [nuevavariable,nuevalista]

